<?php
namespace MetatickPluginLib\Model;

class TableModel extends Model {

    /*
     * fields is an associative array in the form of:
     * field name => field type
     * where field type is one of:
     * int
     * bool
     * float
     * string
     */
    public static $fields = [];

    public static $operators = [
        '=',
        '<>',
        '>',
        '>=',
        '<',
        '<=',
        'LIKE',
        'NOT LIKE',
        'BETWEEN',
        'NOT BETWEEN',
        'IS',
        'IS NOT',
        'IN',
        'NOT IN'
    ];

    public static $network = false;
    public static $table = null;
    public static $key_field = 'id';

    public $is_new = true;

    public static function get_current_version(){
        $option = static::$table . "_table_version";
        if(static::$network){
            return get_site_option($option, -1);
        }
        return get_option($option, -1);
    }

    public static function set_current_version($version){
        $option = static::$table . "_table_version";
        if(static::$network){
            return update_site_option($option, $version);
        }
        return update_option($option, $version);
    }

    public static function get_table_name(){
        global $wpdb;
        if(static::$network){
            $prefix = $wpdb->get_blog_prefix(0);
        }else{
            $prefix = $wpdb->get_blog_prefix();
        }

        return $prefix . static::$table;
    }

    public static function get_charset_collate(){
        global $wpdb;
        return $wpdb->get_charset_collate();
    }

    public static function migrate($folder){
        $files = scandir($folder);
        $version = static::get_current_version();
        foreach($files as $file){
            $pattern = '/^([A-Za-z0-9_]+)\-([0-9]+)\.sql$/';
            if(preg_match($pattern, $file, $matches) === 1){
                if(static::$table !== $matches[1]){
                    continue;
                }
                if($version >= $matches[2]){
                    continue;
                }
                $path = realpath($folder . DIRECTORY_SEPARATOR . $file);
                static::apply_migration($path, $matches[2]);
            }
        }
    }

    public static function apply_migration($path, $version){
        $search = [
            '%%table%%',
            '%%charset_collate%%'
        ];

        $replace = [
            static::get_table_name(),
            static::get_charset_collate()
        ];

        $template = file_get_contents($path);
        $sql = str_replace($search, $replace, $template);
        static::wpdb_query($sql);
        static::set_current_version($version);
    }

    public static function where($where){

        if(is_string($where)){
            /*
             * This is freeform (obviously), but is intended for use with SQL functions like MAX, MIN, etc..
             * this can be nested inside an array, see below
             */
            return $where;
        }
        if(is_array($where)){
            /*
             * All where arrays must be in the following two formats:
             * [field, relation, data]
             * [AND|OR, clause 1, clause 2, clause n]
             *
             * Both the first and second formats can be nested within the second type
             */

            if(in_array($where[0], ['AND', 'OR'])) {
                $operator = $where[0];
                if(count($where) < 2){
                    throw new \Exception("Invalid WHERE clause, $operator requires at least 1 clause");
                }

                $clauses = [];
                foreach (array_slice($where, 1) as $clause) {
                    $clauses[] = static::where($clause);
                }

                return "(\n" . implode("\n$operator ", $clauses) . "\n)";
            }else{
                if(count($where) != 3){
                    throw new \Exception("Invalid WHERE clause");
                }
                list($field, $operator, $data) = $where;

                if(!in_array($operator, array_keys(static::$operators))){
                    throw new \Exception("Invalid WHERE clause, '$operator' operator unknown");
                }

                switch($operator){
                    case 'BETWEEN':
                    case 'NOT BETWEEN':
                        if(is_array($data) && count($data) == 2){
                            $first = esc_sql($data[0]);
                            $last = esc_sql($data[1]);
                            return "$field $operator '$first' AND '$last'";
                        }else{
                            throw new \Exception("Invalid WHERE clause, $operator clause requires two items");
                        }
                        break;

                    case 'IN':
                    case 'NOT IN':
                        if(is_array($data)) {
                            if (count($data) >= 1) {
                                $items = esc_sql($data);
                                return "$field $operator ('" . implode("', '", $items) . "')";
                            } else {
                                throw new \Exception("Invalid WHERE clause, $operator clause requires at least one item");
                            }
                        }elseif(is_string($data)){
                            return "$field $operator ($data)";
                        }
                        break;
                    case 'IS':
                    case 'IS NOT':
                        if($data == null){
                            return "$field $operator NULL";
                        }
                        $data = esc_sql($data);
                        return "$field $operator '$data'";
                        break;
                    default:
                        $data = esc_sql($data);
                        return "$field $operator '$data'";

                }
            }
        }
        throw new \Exception("Invalid WHERE clause");
    }

    public static function order_by($fields){
        return implode(", ", $fields);
    }

    public static function group_by($fields){
        return implode(", ", $fields);
    }

    public static function select($options = []){
        $defaults = [
            'fields' => null,
            'where' => null,
            'having' => null,
            'page' => null,
            'page_size' => null,
            'order_by' => null,
            'group_by' => null,
            'distinct' => false
        ];

        $arguments = array_merge($defaults, $options);

        $fields = $arguments['fields'];
        $where = $arguments['where'];
        $having = $arguments['having'];
        $page = $arguments['page'];
        $page_size = $arguments['page_size'];
        $order_by = $arguments['order_by'];
        $group_by = $arguments['group_by'];
        $distinct = $arguments['distinct'];


        if($fields == null){
            $fields = array_keys(static::$fields);
        }

        $distinct = ($distinct ? 'DISTINCT ' : '');

        $sql = "SELECT $distinct\n" .
            implode(", ", $fields) . "\n" .
            'FROM ' . static::get_table_name();

        if($where != null){
            $sql .= "\nWHERE " . static::where($where);
        }

        if($group_by != null){
            $sql .= "\nGROUP BY " . static::group_by($group_by);
        }

        if($having != null){
            $sql .= "\nHAVING " . static::where($having);
        }

        if($order_by != null){
            $sql .= "\nORDER BY " . static::order_by($order_by);
        }

        if($page !== null && $page_size !== null){
            $offset = $page * $page_size;
            $sql .= "\nLIMIT $offset, $page_size";
        }
        return $sql;
    }

    public static function filter($options = []){
        $sql = static::select($options);

        $results = [];
        foreach(static::wpdb_get_results($sql) as $row){
            $record = new self();
            $record->from_array($row);
            $results[] = $record;
        }
        return $results;
    }

    public static function get($options = []){
        $results = static::filter($options);
        if(count($results) == 0){
            return null;
        }
        return $results[0];
    }

    public static function exists($options = []){
        $results = static::filter($options);
        return count($results) > 0;
    }

    public static function wpdb_query($sql){
        global $wpdb;
        $show_errors = $wpdb->hide_errors();
        $result = $wpdb->query($sql);
        if($show_errors){
            $wpdb->show_errors();
        }
        if($wpdb->last_error !== ''){
            throw new \Exception($wpdb->last_error);
        }
        return $result;
    }

    public static function wpdb_get_results($sql){
        global $wpdb;
        $show_errors = $wpdb->hide_errors();
        $result = $wpdb->get_results($sql, ARRAY_A);
        if($show_errors){
            $wpdb->show_errors();
        }
        if($wpdb->last_error !== ''){
            throw new \Exception($wpdb->last_error);
        }
        return $result;
    }

    public static function wpdb_insert($table, $data, $format){
        global $wpdb;
        $show_errors = $wpdb->hide_errors();
        $wpdb->insert($table, $data, $format);
        if($show_errors){
            $wpdb->show_errors();
        }
        if($wpdb->last_error !== ''){
            throw new \Exception($wpdb->last_error);
        }
        return $wpdb->insert_id;
    }

    public static function wpdb_update($table, $data, $where, $format, $where_format){
        global $wpdb;
        $show_errors = $wpdb->hide_errors();
        $result = $wpdb->update($table, $data, $where, $format, $where_format);
        if($show_errors){
            $wpdb->show_errors();
        }
        if($wpdb->last_error !== ''){
            throw new \Exception($wpdb->last_error);
        }
        return $result;
    }

    public static function wpdb_delete($table, $where, $where_format){
        global $wpdb;
        $show_errors = $wpdb->hide_errors();
        $result = $wpdb->delete($table, $where, $where_format);
        if($show_errors){
            $wpdb->show_errors();
        }
        if($wpdb->last_error !== ''){
            throw new \Exception($wpdb->last_error);
        }
        return $result;
    }

    public static function php_type_to_wordpress($type){
        switch($type){
            case 'int':
            case 'bool':
                return '%d';
            case 'float':
                return '%f';
            case 'string':
                return '%s';
        }
        throw new \TypeError('Invalid Type');
    }

    public static function mysql_to_php($data, $type){
        switch($type){
            case 'int':
            case 'float':
            case 'string':
                return $data;
            case 'bool':
                return !!$data;
        }
        throw new \TypeError('Invalid Type');
    }

    public static function php_to_mysql($data, $type){
        switch($type){
            case 'int':
                return (int)$data;
            case 'bool':
                return ($data ? 1 : 0);
            case 'float':
                return (float)$data;
            case 'string':
                return (string)$data;
        }
        throw new \TypeError('Invalid Type');
    }

    public function from_array($data){

        foreach($data as $field => $value){
            if(array_key_exists($field, static::$fields)){
                $this->properties[$field] = static::mysql_to_php($value, static::$fields[$field]);
            }else{
                $this->properties[$field] = $value;
            }
            if($field == static::$key_field){
                $this->is_new = false;
            }
        }
    }

    public function save(){
        if($this->is_new){
            $this->insert();
        }else{
            $this->update();
        }
    }

    public function insert(){
        $table = static::get_table_name();

        $data = [];
        $format = [];
        foreach(static::$fields as $field => $type){
            if($field != static::$key_field){
                $format[] = static::php_type_to_wordpress($type);
                $data[$field] = static::php_to_mysql($this->$field, $type);
            }
        }
        $insert_id = static::wpdb_insert($table, $data, $format);
        $this->properties[static::$key_field] = $insert_id;
    }

    public function update(){
        $table = static::get_table_name();

        $where = [
            static::$key_field => $this->properties[static::$key_field]
        ];
        $where_format = [
            static::php_type_to_wordpress(static::$fields[static::$key_field])
        ];

        $data = [];
        $format = [];
        foreach(static::$fields as $field => $type){
            if($field != static::$key_field){
                $format[] = static::php_type_to_wordpress($type);
                $data[$field] = static::php_to_mysql($this->$field, $type);
            }
        }

        static::wpdb_update($table, $data, $where, $format, $where_format);
    }

    public function delete(){
        $table = static::get_table_name();

        $where = [
            static::$key_field => $this->properties[static::$key_field]
        ];
        $where_format = [
            static::php_type_to_wordpress(static::$fields[static::$key_field])
        ];

        static::wpdb_delete($table, $where, $where_format);
    }
}